import torch
from torch import nn
import numpy as np
from torch.nn.functional import one_hot
import lips
from utils import get_all_replacements

class ConvModel(nn.Module):
    def __init__(self, n_chars, embed_size=256, hidden_size=256, n_classes=10, kernel_size = 5, n_layers = 2,p=2, reduce = 'mean', lips_emb = True):
        super(ConvModel, self).__init__()
        self.embed_size = embed_size
        self.hidden_size = hidden_size
        self.n_classes = n_classes
        self.n_chars = n_chars
        self.kernel_size = kernel_size
        self.emb = nn.Embedding(n_chars, embed_size,padding_idx=0)
        self.n_layers = n_layers
        self.reduce = reduce
        self.lips_emb = lips_emb
        for i in range(self.n_layers):
            if i==0:
                setattr(self,f'conv{i}',nn.Conv1d(embed_size,hidden_size,kernel_size=self.kernel_size,stride=1, bias=False))
            else:
                setattr(self,f'conv{i}',nn.Conv1d(hidden_size,hidden_size,kernel_size=self.kernel_size,stride=1, bias=False))
        self.act = nn.ReLU()
        self.C = nn.Linear(self.hidden_size, self.n_classes, bias=False)
        self.p=p
        if self.p == 'inf':
            self.p = float(self.p)
        self.init_1lips()

    def init_1lips(self):
        with torch.no_grad():
            self.emb.weight/=lips.M_emb(self.emb,p=self.p)

            for i in range(self.n_layers):
                getattr(self,f'conv{i}').weight/=lips.M_conv(getattr(self,f'conv{i}'),p=self.p)

            if self.p==1:
                q = float('inf')
            elif self.p==2:
                q = 2
            elif self.p==float('inf'):
                q = 1

            self.C.weight/=lips.M_emb(self.C,p=q)
        
        
    def compute_lips(self,x, target=0, label=1):
        '''
        compute the local lipschitz constant evaluated at x, with max levenshtein distance k
        '''
        with torch.no_grad():
            if self.lips_emb:
                ME = lips.M_emb(self.emb,x=x,p=self.p)
            else:
                ME = 1
            MK = 1
            for i in range(self.n_layers):
                MK *=lips.M_conv(getattr(self,f'conv{i}'),p=self.p)
            
        

            if self.p==1:
                q = float('inf')
            elif self.p==2:
                q = 2
            elif self.p==float('inf'):
                q = 1
            if self.reduce == 'sum':
                return (torch.norm(self.C.weight[target] - self.C.weight[label],p=q)*MK*ME)
            elif self.reduce == 'mean':
                return (torch.norm(self.C.weight[target] - self.C.weight[label],p=q)*MK*ME)*2/x.shape[1]
    
    def compute_lips_reg(self,x, target=0, label=1):
        '''
        compute the local lipschitz constant evaluated at x, with max levenshtein distance k
        '''
        if self.lips_emb:
            ME = lips.M_emb(self.emb,p=self.p)
        else:
            ME = 1
        
        MK = 1
        for i in range(self.n_layers):
            MK *=lips.M_conv(getattr(self,f'conv{i}'),p=self.p)

        if self.p==1:
            q = float('inf')
        elif self.p==2:
            q = 2
        elif self.p==float('inf'):
            q = 1

        #print(ME, MK, torch.norm(self.C.weight[0] - self.C.weight[1],p=q))
        #print('W',lips.M_emb(self.C,p=q).item(),'K',MK.item(),'E',ME.item())
        return (lips.M_emb(self.C,p=q)*MK*ME)

    def forward(self, x, mask=None):
        bs, length = x.shape[0], x.shape[1]
        if mask is None:
            mask = torch.ones([bs,length,1],device=x.device)
        x = (mask*self.emb(x)).transpose(1,2)
        for i in range(self.n_layers):
            x = self.act(getattr(self,f'conv{i}')(x))
        x = x.transpose(1,2)

        if self.reduce == 'sum':
            x = x.sum(dim=1)
        elif self.reduce == 'mean':
            x = x.mean(dim=1)
        out = self.C(x)
        return out, mask.unsqueeze(-1)
    
    def forward_bounds_till_reduce(self, x, X, delta):
        z = (self.emb(x) + delta*(self.emb(X)-self.emb(x))).transpose(1,2)
        for i in range(self.n_layers):
            z = self.act(getattr(self,f'conv{i}')(z))
        z = z.transpose(1,2)

        if self.reduce == 'sum':
            z = z.sum(dim=1)
        elif self.reduce == 'mean':
            z = z.mean(dim=1)

        return z.min(axis=0,keepdim=True)[0],z.max(axis=0,keepdim=True)[0]
    
    def bounds_logits(self,l,u,y):
        Cp = ((self.C.weight > 0)*self.C.weight).transpose(0,1)
        Cn = ((self.C.weight < 0)*self.C.weight).transpose(0,1)

        L = l@Cp + u@Cn
        U = u@Cp + l@Cn
        y_one_hot = one_hot(y,num_classes=self.n_classes)
        #print(L,U)
        return L*y_one_hot + U*(1-y_one_hot)

class ConvLipsModel(nn.Module):
    def __init__(self, n_chars, embed_size=256, hidden_size=256, n_classes=10, kernel_size = 5, n_layers = 2,p=2, reduce = 'mean', lips_emb = True, batch_norm = False):
        super(ConvLipsModel, self).__init__()
        self.embed_size = embed_size
        self.hidden_size = hidden_size
        self.n_classes = n_classes
        self.n_chars = n_chars
        self.kernel_size = kernel_size
        self.emb = nn.Embedding(n_chars, embed_size,padding_idx=0)
        self.n_layers = n_layers
        self.reduce = reduce
        self.lips_emb = lips_emb
        self.batch_norm = batch_norm
        for i in range(self.n_layers):
            if self.batch_norm:
                setattr(self,f'bn{i}', nn.BatchNorm1d(hidden_size, affine=False))
            if i==0:
                setattr(self,f'conv{i}',nn.Conv1d(embed_size,hidden_size,kernel_size=self.kernel_size,stride=1, bias=False))
            else:
                setattr(self,f'conv{i}',nn.Conv1d(hidden_size,hidden_size,kernel_size=self.kernel_size,stride=1, bias=False))
        self.act = nn.ReLU()
        self.C = nn.Linear(self.hidden_size, self.n_classes, bias=False)
        self.p=p
        self.p = float(self.p)

    def compute_lips(self,x, target=0, label=1):
        '''
        compute the local lipschitz constant evaluated at x, with max levenshtein distance k
        '''
        with torch.no_grad():
            if self.p==1:
                q = float('inf')
            elif self.p==2:
                q = 2
            elif self.p==float('inf'):
                q = 1
            L = 1
            # if self.batch_norm:
            #     for i in range(self.n_layers):
            #         L*=torch.max(1/torch.sqrt(getattr(self,f'bn{i}').running_var + getattr(self,f'bn{i}').eps))
            if self.reduce == 'sum':
                if self.lips_emb:
                    return L*(lips.M_emb(self.emb,x=x,p=self.p)/lips.M_emb(self.emb,p=self.p))*torch.norm(self.C.weight[target] - self.C.weight[label],p=q)/lips.M_emb(self.C,p=q)

                else:
                    return L*torch.norm(self.C.weight[target] - self.C.weight[label],p=q)/lips.M_emb(self.C,p=q)*(1/torch.max(torch.linalg.norm(self.emb.weight.data,dim=1,ord=self.p)))

            elif self.reduce == 'mean':
                return L*2/x.shape[1]*(lips.M_emb(self.emb,x=x,p=self.p)/lips.M_emb(self.emb,p=self.p))*torch.norm(self.C.weight[target] - self.C.weight[label],p=q)/lips.M_emb(self.C,p=q)
    
    def compute_lips_reg(self,x, target=0, label=1):
        '''
        compute the local lipschitz constant evaluated at x, with max levenshtein distance k
        '''
        return 1

    def forward(self, x, mask=None):
        bs, length = x.shape[0], x.shape[1]
        if mask is None:
            mask = torch.ones([bs,length,1],device=x.device)
        if self.lips_emb:
            ME = lips.M_emb(self.emb,p=self.p)
        else:
            ME = torch.max(torch.linalg.norm(self.emb.weight.data,dim=1,ord=self.p))
            #print(ME)
        x = (mask*self.emb(x)).transpose(1,2)/ME
        for i in range(self.n_layers):
            #print(i,x.shape)
            MK =lips.M_conv(getattr(self,f'conv{i}'),p=self.p)
            x = getattr(self,f'conv{i}')(x/(MK))
            if self.batch_norm:
                MB = torch.max(1/torch.sqrt(getattr(self,f'bn{i}').running_var + getattr(self,f'bn{i}').eps))
                x = getattr(self,f'bn{i}')(x/MB)
            x = self.act(x)
        x = x.transpose(1,2)

        #x, _ = x.max(dim=1)
        if self.reduce == 'sum':
            x = x.sum(dim=1)
        elif self.reduce == 'mean':
            x = x.mean(dim=1)
        if self.p==1:
            q = float('inf')
        elif self.p==2:
            q = 2
        elif self.p==float('inf'):
            q = 1
        MC = lips.M_emb(self.C,p=q)
        out = self.C(x/MC)
        return out, mask.unsqueeze(-1)
    
    def forward_bounds_till_reduce(self, x, X, delta):
        if self.lips_emb:
            ME = lips.M_emb(self.emb,p=self.p)
        else:
            ME = torch.max(torch.linalg.norm(self.emb.weight,dim=1,p=self.p))
        z = (self.emb(x) + delta*(self.emb(X)-self.emb(x))).transpose(1,2)/ME
        for i in range(self.n_layers):
            MK =lips.M_conv(getattr(self,f'conv{i}'),p=self.p)
            z = self.act(getattr(self,f'conv{i}')(z/MK))
        z = z.transpose(1,2)

        if self.reduce == 'sum':
            z = z.sum(dim=1)
        elif self.reduce == 'mean':
            z = z.mean(dim=1)

        return z.min(axis=0,keepdim=True)[0],z.max(axis=0,keepdim=True)[0]
    
    def bounds_logits(self,l,u,y):
        Cp = ((self.C.weight > 0)*self.C.weight).transpose(0,1)
        Cn = ((self.C.weight < 0)*self.C.weight).transpose(0,1)

        if self.p==1:
            q = float('inf')
        elif self.p==2:
            q = 2
        elif self.p==float('inf'):
            q = 1
        MC = lips.M_emb(self.C,p=q)
        L = (l@Cp + u@Cn)/MC
        U = (u@Cp + l@Cn)/MC
        y_one_hot = one_hot(y,num_classes=self.n_classes)
        return L*y_one_hot + U*(1-y_one_hot)
    
if __name__ == '__main__':
    model = ConvModel(4, embed_size=150, hidden_size=10, n_classes=2, kernel_size = 5, n_layers = 1,p='inf', reduce = 'mean')
    sentence = 'happy'
    char_to_id = {'h':0, 'a':1, 'p': 2, 'y':3}
    S = get_all_replacements(sentence,char_to_id)
    y = torch.tensor([1])
    print(S)
    x = torch.tensor([char_to_id[c] for c in sentence]).unsqueeze(0)
    X = torch.tensor([[char_to_id[c] for c in s] for s in S])
    print(x.shape,X.shape)
    l,u = model.forward_bounds_till_reduce(x,X,delta=1)
    print(l,u)

    y_ = model.bounds_logits(l,u,y)
    print(y_)